# Generated by Django 3.2.12 on 2023-05-19 07:09

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0003_create_workspace'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='token',
            name='owner',
        ),
        migrations.AddField(
            model_name='token',
            name='user',
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to=settings.AUTH_USER_MODEL,
            ),
        ),
    ]
