.order-active {
    color: inherit;
    text-decoration: none; /* avoid link underline */
}

.order-inactive {
    color: grey;
    text-decoration: none; /* avoid link underline */
}

.errorlist {
    color:red;
}

footer {
    font-size: 80%;
}
footer a {
    text-decoration: none;
}
footer a:visited {
}

#beta-stamp {
 position:fixed;
 bottom:5px;
 right:5px;
 opacity:0.5;
 z-index:99;
 color:red;
 font-family: monospace;
 font-size: 3rem;
 font-weight: bold;
 rotate: -30deg;
}

pre.file_contents_lineno a {
    text-decoration: none;
    color: inherit;
}
pre.file_contents_lineno a:hover {
    text-decoration: underline;
}

div.file_highlighted pre {
    margin: 0;
}

table.wr-artifacts tbody:before {
    content: '';
    margin-top: 1em;
    display: block;
}

/* Style logout button to appear in the same way it did when we were logging
 * out using a GET request via a link.
 */
#logout-form button {
    background: none;
    border: none;
    cursor: pointer;
    padding: 0;
}

/*
 * A height attribute (any height attribute) is needed on a table element to be
 * able to use height: 100% inside table cells.
 *
 * See https://stackoverflow.com/a/73399942
 */
table {
    height: fit-content;
}


form.table-filter-form .select2
{
    flex-grow: 1;
}
