# Generated by Django 4.2.15 on 2024-12-28 23:35

import django.core.serializers.json
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0112_workrequest_autopkgtest_extra_repositories'),
    ]

    operations = [
        migrations.AddField(
            model_name='workrequest',
            name='output_data_json',
            field=models.JSONField(
                blank=True,
                db_column='output_data',
                encoder=django.core.serializers.json.DjangoJSONEncoder,
                null=True,
            ),
        ),
    ]
