# Generated by Django 4.2.15 on 2024-11-01 10:05

from django.db import migrations, models

import debusine.db.models.workspaces


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0095_create_singleton_collections'),
    ]

    operations = [
        migrations.AlterField(
            model_name='workspace',
            name='name',
            field=models.CharField(
                max_length=255,
                validators=[
                    debusine.db.models.workspaces.validate_workspace_name
                ],
            ),
        ),
    ]
