# Generated by Django 3.2.19 on 2023-12-14 09:10

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0011_auto_20231024_1213'),
    ]

    operations = [
        migrations.AddField(
            model_name='workspace',
            name='default_expiration_delay',
            field=models.PositiveIntegerField(
                default=0,
                help_text='minimal time (in days) that a new artifact is kept in the workspace before being expired',
            ),
        ),
    ]
