# Generated by Django 3.2.19 on 2024-09-19 08:27

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('db', '0081_create_fallback_scope'),
    ]

    operations = [
        migrations.AddField(
            model_name='workspace',
            name='scope',
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name='workspaces',
                to='db.scope',
            ),
        ),
    ]
