# Generated by Django 4.2.18 on 2025-02-06 14:43

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('db', '0136_add_groupmembership'),
    ]

    operations = [
        migrations.AddField(
            model_name='groupmembership',
            name='role',
            field=models.CharField(
                choices=[('admin', 'Admin'), ('member', 'Member')],
                default='member',
                max_length=16,
            ),
        ),
    ]
