# Generated by Django 3.2.19 on 2024-04-16 16:11

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('db', '0033_collectionitem_tighten_unique_active_name'),
    ]

    operations = [
        migrations.RenameField(
            model_name='workrequest',
            old_name='event_reactions',
            new_name='event_reactions_json',
        ),
        migrations.RenameField(
            model_name='workrequest',
            old_name='workflow_data',
            new_name='workflow_data_json',
        ),
        migrations.AlterField(
            model_name='workrequest',
            name='event_reactions_json',
            field=models.JSONField(
                blank=True, db_column='event_reactions', default=dict
            ),
        ),
        migrations.AlterField(
            model_name='workrequest',
            name='workflow_data_json',
            field=models.JSONField(
                blank=True, db_column='workflow_data', default=dict
            ),
        ),
    ]
