# Generated by Django 3.2.19 on 2024-05-03 15:30

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('db', '0047_sbuild_lookup'),
    ]

    operations = [
        migrations.AlterField(
            model_name='collection',
            name='retains_artifacts',
            field=models.BooleanField(default=True),
        ),
    ]
