# Generated by Django 4.2.15 on 2024-12-09 22:10

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0102_collection_related_name'),
    ]

    operations = [
        migrations.AlterField(
            model_name='scoperole',
            name='group',
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name='scope_roles',
                to='db.group',
            ),
        ),
        migrations.AlterField(
            model_name='scoperole',
            name='resource',
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name='roles',
                to='db.scope',
            ),
        ),
        migrations.AlterField(
            model_name='workspacerole',
            name='group',
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name='workspace_roles',
                to='db.group',
            ),
        ),
        migrations.AlterField(
            model_name='workspacerole',
            name='resource',
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name='roles',
                to='db.workspace',
            ),
        ),
    ]
