# Generated by Django 4.2.18 on 2025-02-13 18:06

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0131_workrequest_add_configured_task_data_and_version'),
    ]

    # 0128_filestore_add_size_columns was retroactively hacked to add these
    # as BigIntegerField rather than IntegerField in order that
    # 0129_filestore_total_size can be applied on instances where there are
    # file stores of size 2 GiB or more, but we also alter them here for the
    # benefit of systems that had already successfully applied those
    # migrations.
    operations = [
        migrations.AlterField(
            model_name='filestore',
            name='max_size',
            field=models.BigIntegerField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='filestore',
            name='soft_max_size',
            field=models.BigIntegerField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='filestore',
            name='total_size',
            field=models.BigIntegerField(default=0, editable=False),
        ),
    ]
