# Generated by Django 4.2.15 on 2024-12-16 10:09

from django.db import migrations, models

import debusine.db.models.scopes


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0106_workrequest_workflow_template'),
    ]

    operations = [
        migrations.AddField(
            model_name='scope',
            name='icon',
            field=models.CharField(
                blank=True,
                default='',
                help_text='Optional user-visible icon, resolved via ``{% static %}`` in templates',
                max_length=255,
            ),
        ),
        migrations.AddField(
            model_name='scope',
            name='label',
            field=models.CharField(
                blank=True,
                default='',
                help_text='User-visible name for the scope',
                max_length=255,
            ),
        ),
        migrations.AlterField(
            model_name='scope',
            name='name',
            field=models.CharField(
                help_text='internal name for the scope',
                max_length=255,
                unique=True,
                validators=[debusine.db.models.scopes.validate_scope_name],
            ),
        ),
    ]
