# Generated by Django 3.2.19 on 2023-09-08 11:21

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0008_artifact_created_by'),
    ]

    operations = [
        migrations.AlterField(
            model_name='token',
            name='comment',
            field=models.CharField(
                blank=True,
                default='',
                max_length=100,
                verbose_name='Reason that this token was created',
            ),
        ),
    ]
