# Generated by Django 4.2.15 on 2024-12-05 13:43

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0101_collectionitem_suite_indexes'),
    ]

    operations = [
        migrations.AlterField(
            model_name='collection',
            name='workspace',
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name='collections',
                to='db.workspace',
            ),
        ),
    ]
