# Generated by Django 4.2.18 on 2025-02-03 20:18

import django.core.serializers.json
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('db', '0130_filestore_total_size_triggers'),
    ]

    operations = [
        migrations.AddField(
            model_name='workrequest',
            name='configured_task_data',
            field=models.JSONField(
                blank=True,
                encoder=django.core.serializers.json.DjangoJSONEncoder,
                help_text='task_data with task configuration applied',
                null=True,
            ),
        ),
        migrations.AddField(
            model_name='workrequest',
            name='version',
            field=models.IntegerField(
                default=0,
                help_text='version of the code that computed the dynamic task data',
            ),
        ),
    ]
