# Generated by Django 4.2.15 on 2024-11-06 00:32

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0097_recreate_json_data_unique_constraint'),
    ]

    operations = [
        migrations.AddField(
            model_name='artifact',
            name='original_artifact',
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to='db.artifact',
            ),
        ),
    ]
