# Generated by Django 4.2.18 on 2025-02-25 17:37

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0133_filestore_provider_account_alter_filestore_backend'),
    ]

    operations = [
        migrations.AddConstraint(
            model_name='filestoreinscope',
            constraint=models.CheckConstraint(
                check=models.Q(
                    ('populate', False),
                    models.Q(('drain', False), ('read_only', False)),
                    _connector='OR',
                ),
                name='db_filestoreinscope_consistent_populate',
            ),
        ),
    ]
