# Generated by Django 4.2.18 on 2025-02-11 15:50

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0127_workspace_add_expiry_fields'),
    ]

    operations = [
        migrations.AddField(
            model_name='filestore',
            name='max_size',
            field=models.BigIntegerField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='filestore',
            name='soft_max_size',
            field=models.BigIntegerField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='filestore',
            name='total_size',
            field=models.BigIntegerField(blank=True, null=True, editable=False),
        ),
        migrations.AddConstraint(
            model_name='filestore',
            constraint=models.CheckConstraint(
                check=models.Q(
                    ('soft_max_size__gt', models.F('max_size')), _negated=True
                ),
                name='db_filestore_consistent_max_sizes',
            ),
        ),
    ]
