# Generated by Django 3.2.19 on 2024-05-07 17:01

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('db', '0049_update_suite_lintian_collection_lookup'),
    ]

    operations = [
        migrations.RenameField(
            model_name='collection',
            old_name='retains_artifacts',
            new_name='retains_artifacts_bool',
        ),
        migrations.AddField(
            model_name='collection',
            name='retains_artifacts',
            field=models.CharField(
                choices=[
                    ('never', 'Never'),
                    ('workflow', 'While workflow is running'),
                    ('always', 'Always'),
                ],
                default='always',
                max_length=8,
            ),
        ),
        migrations.AddField(
            model_name='collection',
            name='workflow',
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to='db.workrequest',
            ),
        ),
    ]
