# Generated by Django 3.2.19 on 2024-03-25 13:28

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('db', '0026_suite_collection'),
    ]

    operations = [
        migrations.AddField(
            model_name='workrequest',
            name='task_type',
            field=models.CharField(
                choices=[
                    ('worker', 'Worker'),
                    ('server', 'Server'),
                    ('internal', 'Internal'),
                    ('workflow', 'Workflow'),
                ],
                default='worker',
                editable=False,
                max_length=8,
                verbose_name='Type of task to execute',
            ),
        ),
    ]
